Add LoadPath "CoqEAL".
Add LoadPath "homology".
Add LoadPath "persistent_homology".

Require Import ssreflect ssrbool ssrfun eqtype ssrnat seq div choice. 
Require Import fintype finfun bigop finset ssralg perm zmodp matrix.
Require Import persistentincidencematrices incidencematrices homology.

Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.

Open Scope ring_scope.

Section example1.

(* First example: a filled triangle. In the filtration, we
firstly add each one of the vertices, then the edges and finally
the triangle *)

Definition f1 : seq (seq (seq 'I_3)) := 
[::[::[::0]];
   [::[::0];[::1]];
   [::[::0];[::1];[::2%:R]];
   [::[::0;1];[::2%:R]];
   [::[::0;1];[::1;2%:R]];
   [::[::0;1];[::1;2%:R];[::0;2%:R]];
   [::[::0;1;2%:R]]].

(* A connected component which lives at level 0 is still alive at level 1 *)

Eval vm_compute in ex_p_persistent_n_betti_K_j f1 0 0 1.


(* Two connected components which live at level 1 are still alive at 
   level 2 *)

Eval vm_compute in ex_p_persistent_n_betti_K_j f1 0 1 2.

(* The hole at level 5 dies entering level 6 *)

Eval vm_compute in ex_Betti_sc (nth nil f1 5) 1.
Eval vm_compute in ex_p_persistent_n_betti_K_j  f1 1 5 6.

End example1.

Section example2.

Let a := 0%:R : 'I_4.
Let b := 1%:R : 'I_4.
Let c := 2%:R : 'I_4.
Let d := 3%:R : 'I_4.

(* Figure 1 of the paper computing persistent homology *)

Definition f2 : seq (seq (seq 'I_4)) := 
[::[::[::a];[::b]];
   [::[::a;b];[::b;c];[::d]];
   [::[::a;b];[::b;c];[::c;d];[::a;d]];
   [::[::a;b];[::b;c];[::c;d];[::a;d];[::c;d]];
   [::[::c;d];[::a;d];[::a;b;c]];
   [::[::a;b;c];[::a;c;d]]].

(* One of the connected component which lives at level 0 dies at level 1
   the other is still alive *)

Eval vm_compute in ex_Betti_sc (nth nil f2 0) 0.
Eval vm_compute in ex_p_persistent_n_betti_K_j  f2 0 0 1.

(* One of the holes which lives at level 3 dies at level 4
   the other is still alive ...*)

Eval vm_compute in ex_p_persistent_n_betti_K_j  f2 1 3 4.

(* ... but dies entering level 5  *)

Eval vm_compute in ex_p_persistent_n_betti_K_j  f2 1 3 5.

End example2.

Section example3.

(* The diabolo example of the paper *)
Local Notation "0" := (0%:R : 'I_6).
Local Notation "1" := (1%:R : 'I_6).
Local Notation "2" := (2%:R : 'I_6).
Local Notation "3" := (3%:R : 'I_6).
Local Notation "4" := (4%:R : 'I_6).
Local Notation "5" := (5%:R : 'I_6).

Definition diabolo_facets : seq (seq 'I_6) :=
  [::[::2;3];[::3;4];[::3;5];[::4;5];[::0;1;2]]. 

Eval vm_compute in ex_Betti_sc diabolo_facets 0.
Eval vm_compute in ex_Betti_sc diabolo_facets 1.

Definition diabolo_filtration : seq (seq (seq 'I_6)) := 
[::[::[::0];[::1];[::2]];
   [::[::0;1];[::1;2];[::0;2]];
   [::[::0;1];[::1;2];[::0;2];[::3];[::4];[::5]];
   [::[::0;1];[::1;2];[::0;2];[::3;4];[::4;5];[::3;5]];
   [::[::0;1];[::1;2];[::0;2];[::3;4];[::4;5];[::3;5];[::2;3]];
   [::[::0;1;2];[::3;4];[::4;5];[::3;5];[::2;3]]].

(* Just one of the connected components of level 0 is still alive
   at level 1 *)

Eval vm_compute in ex_Betti_sc (nth nil diabolo_filtration 0) 0.
Eval vm_compute in ex_p_persistent_n_betti_K_j  diabolo_filtration 0 0 1.

(* The hole which appears at level 1 is still alive at levels 2, 3, 4 
   but dies entering level 5*)

Eval vm_compute in ex_p_persistent_n_betti_K_j diabolo_filtration 1 1 2.
Eval vm_compute in ex_p_persistent_n_betti_K_j diabolo_filtration 1 1 3.
Eval vm_compute in ex_p_persistent_n_betti_K_j diabolo_filtration 1 1 4.
Eval vm_compute in ex_p_persistent_n_betti_K_j diabolo_filtration 1 1 5.

(* The two holes which are live at level 3 are still alive at level 4 *)

Eval vm_compute in ex_Betti_sc (nth nil diabolo_filtration 3) 1.
Eval vm_compute in ex_p_persistent_n_betti_K_j diabolo_filtration 1 3 4.

End example3.

Section generation.

Variable n : nat.
Hypothesis n0 : 0 < n.

Fixpoint vertices_aux (p : nat) :=
match p with
| 0 => nil
| S m => (vertices_aux m) ++ [::[::(cast_ord (prednK n0) (@inZp n.-1 m))]] 
end.

Definition vertices := vertices_aux n.-1.

Fixpoint triangles_aux (p : nat) :=
match p with
| 0 => nil
| 1 => nil
| 2 => nil
| S m => (triangles_aux m) ++ [::[::(cast_ord (prednK n0) (@inZp n.-1 m.-2));
                                    (cast_ord (prednK n0) (@inZp n.-1 m.-1));
                                    (cast_ord (prednK n0) (@inZp n.-1 m))]]
end.

Definition triangles := triangles_aux n.-1.

End generation.

Section triangles.

Lemma h10 : 0 < 10. by done. Qed.
Eval vm_compute in (triangles h10).
Time Eval vm_compute in ex_Betti_sc (triangles h10) 1.
Definition f10 := [::(drop 0 (triangles h10))] ++ [::(triangles h10)].
Time Eval vm_compute in ex_p_persistent_n_betti_K_j  f10 0 0 1.

Lemma h50 : 0 < 50. by done. Qed.
Eval vm_compute in (triangles h50).
Time Eval vm_compute in ex_Betti_sc (triangles h50) 1.
Definition f50 := [::(drop 0 (triangles h50))] ++ [::(triangles h50)].
Time Eval vm_compute in ex_p_persistent_n_betti_K_j  f50 0 0 1.

Lemma h100 : 0 < 100. by done. Qed.
Eval vm_compute in (triangles h100).
Time Eval vm_compute in ex_Betti_sc (triangles h100) 1.
Definition f100 := [::(drop 0 (triangles h100))] ++ [::(triangles h100)].
Time Eval vm_compute in ex_p_persistent_n_betti_K_j  f100 0 0 1.

Lemma h200 : 0 < 200. by done. Qed.
Eval vm_compute in (triangles h200).
Time Eval vm_compute in ex_Betti_sc (triangles h200) 1.
Definition f200 := [::(drop 0 (triangles h200))] ++ [::(triangles h200)].
Time Eval vm_compute in ex_p_persistent_n_betti_K_j  f200 0 0 1.

Lemma h500 : 0 < 500. by done. Qed.
Eval vm_compute in (triangles h500).
Time Eval vm_compute in ex_Betti_sc (triangles h500) 1.
Definition f500 := [::(drop 0 (triangles h500))] ++ [::(triangles h500)].
Time Eval vm_compute in ex_p_persistent_n_betti_K_j  f500 0 0 1.


Section a_lot_of_vertices.


(* 99 vertices and then 10 vertices *)
Lemma h10 : 0 < 10. by done. Qed.
Definition f10 := [::(drop 0 (vertices h10))] ++ [::(vertices h10)].
Time Eval vm_compute in ex_p_persistent_n_betti_K_j  f10 0 0 1.

(* 99 vertices and then 100 vertices (6 seconds) *)
Lemma h100 : 0 < 100. by done. Qed.
Definition f100 := [::(drop 0 (vertices h100))] ++ [::(vertices h100)].
Time Eval vm_compute in ex_p_persistent_n_betti_K_j  f100 0 0 1.

(* 499 vertices and then 500 vertices (13 minutes and 51 seconds ) *)
Lemma h500 : 0 < 500. by done. Qed.
Definition f500 := [::(drop 0 (vertices h500))] ++ [::(vertices h500)].
Time Eval vm_compute in f500.
Time Eval vm_compute in ex_p_persistent_n_betti_K_j f500 0 0 1.

End a_lot_of_vertices.