(*****************************************************************************)
(*                               Homology Groups                             *)
(*****************************************************************************)
Require Import ssreflect ssrfun ssrbool eqtype ssrnat seq choice fintype bigop.
Require Import finfun tuple ssralg matrix mxalgebra zmodp vector seqmatrix rank cssralg.

Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.

(*****************************************************************************

Definition of homology groups:

Let V1, V2 and V3 be vector spaces over a field K and f: V1 -> V2, g: V2 -> V3
be linear maps such that gf = 0 then the homology group of (f,g), denoted by
H_(f,g) is the quotient ker(g)/im(f).

*****************************************************************************)

Section HomologyDefinition. 

Variable (F : fieldType) (V1 V2 V3 : vectType F) 
         (f : 'Hom(V1,V2)) (g : 'Hom(V2,V3)).

Definition Homology := ((lker g) :\: (limg f))%VS.

Section HomologyDefinitionProperties.

Lemma img_f_subset_ker_g: (g \o f = 0%R)%VF -> 
   ((limg f) <= (lker g))%VS.
Proof.
by move => nilpotency; rewrite lkerE -limg_comp nilpotency lim0g.
Qed.

End HomologyDefinitionProperties. 

Section HomologyDimension.

Definition Betti := \dim Homology.

Lemma addn_equal_subn: forall (m n p : nat), m+n= p -> m = p - n.
Proof.
move=> m n p H.
have H1: m+n - n = p -n.
  by rewrite H.
by rewrite -H1 addnK.
Qed.

Lemma dim_lker_limg_homology: \dim ((lker g) :\: (limg f)) = \dim (lker g) - \dim ((lker g) :&: (limg f)).
Proof.
have H: \dim ((lker g) :\: (limg f)) + \dim ((lker g) :&: (limg f))= \dim (lker g).
  by rewrite addnC dimv_cap_compl.
by apply addn_equal_subn.
Qed.

Lemma dim_ker_intersection_im: (g \o f = 0%R)%VF -> \dim ((lker g) :&: (limg f)) = \dim (limg f).
Proof.
by move => nilpotency; move :  (img_f_subset_ker_g nilpotency);
move/capv_idPr => ->.
Qed.

Lemma dim_ker : \dim (lker g) = Vector.dim V2 - \dim (limg g).
Proof.
have : \dim ((@fullv _ V2) :&: lker g) + \dim (g @: (@fullv _ V2)) = \dim (@fullv _ V2).
  by rewrite limg_ker_dim.
have H1 : \dim ((@fullv _ V2) :&: lker g) = \dim (lker g).
  by move : (subvf (lker g)) => /capv_idPr ->. 
by rewrite H1 dimvf => H; apply addn_equal_subn.
Qed.

Lemma BettiE: (g \o f = 0%R)%VF ->
  Betti = Vector.dim V2 - \dim (limg g) - \dim (limg f).
Proof.
move => nilpotency.
by rewrite /Betti /Homology dim_lker_limg_homology dim_ker dim_ker_intersection_im.
Qed.

End HomologyDimension.

End HomologyDefinition.

Section HomologyRank.

Variable (K : fieldType) (v1 v2 v3:nat).

Definition V1 := (matrix_vectType K 1 v1).
Definition V2 := (matrix_vectType K 1 v2).
Definition V3 := (matrix_vectType K 1 v3).

Variable (mxf: 'M[K]_(Vector.dim V1,Vector.dim V2)) 
  (mxg : 'M[K]_(Vector.dim V2,Vector.dim V3)).

Definition Betti_rank := v2 - \rank mxg - \rank mxf.

Local Open Scope ring_scope.

Lemma vdimV2: Vector.dim V2 = v2.
Proof.
by rewrite /V2 /Vector.dim /= mul1n.
Qed.

(* Previously called dimHomologyrankE *)
Lemma Betti_rankE: mxf *m mxg = 0 ->
  Betti_rank = Betti (Vector.Hom mxf) (Vector.Hom mxg).
Proof. 
move=> nilpotency_mx; rewrite BettiE.
  by rewrite vdimV2 unlock /dimv !genmxE //= !genmx1 !mul1mx.
apply/lfunP => u. rewrite comp_lfunE !lfunE /fun_of_lfun unlock //=
/fun_of_lfun_def //= Vector.InternalTheory.r2vK -mulmxA
nilpotency_mx mulmx0.
suff : Vector.InternalTheory.v2r 0 = 0.
  by move => <-; rewrite Vector.InternalTheory.v2rK.
move => t t0.
have := s2valP (Vector.InternalTheory.v2r_subproof t0) => H.
rewrite /Vector.InternalTheory.v2r.
by move: (GRing.linear0 (Linear H)).
Qed.

End HomologyRank.

Section Ex_Homology_Def.

Variable (K : fieldType) (v1 v2 v3:nat).
Variable (CK : cunitRingType K).

Hypothesis v1_0 : v1 > 0.
Hypothesis v2_0 : v2 > 0.
Hypothesis v3_0 : v3 > 0.

Definition W1 := (matrix_vectType K 1 v1).
Definition W2 := (matrix_vectType K 1 v2).
Definition W3 := (matrix_vectType K 1 v3).

Lemma vdimW2: Vector.dim W2 = v2.
Proof.
by rewrite /W2 /Vector.dim /= mul1n.
Qed.

Lemma vdimW1: Vector.dim W1 = v1.
Proof.
by rewrite /W1 /Vector.dim /= mul1n.
Qed.

Lemma vdimW3: Vector.dim W3 = v3.
Proof.
by rewrite /W3 /Vector.dim /= mul1n.
Qed.

Lemma Sv1: exists x, 1+x = v1.
Proof. 
by exists (v1-1); rewrite addnC subnK.
Qed.

Lemma Sv2: exists x, 1+x = v2.
Proof. 
by exists (v2-1); rewrite addnC subnK.
Qed.

Lemma Sv3: exists x, 1+x = v3.
Proof. 
by exists (v3-1); rewrite addnC subnK.
Qed.

Definition ex_Betti_rank (mxf mxg : seqmatrix CK) :=
  v2 - (rank_elim_seqmx v2 v3 mxg) - (rank_elim_seqmx v1 v2 mxf).

Lemma ex_Betti_rankE: forall (mxf: 'M[K]_(Vector.dim W1, Vector.dim W2)) (mxg : 'M[K]_(Vector.dim W2, Vector.dim W3)),
   ex_Betti_rank (seqmx_of_mx _ mxf) (seqmx_of_mx _ mxg) = Betti_rank mxf mxg.
Proof. 
rewrite /ex_Betti_rank /Betti_rank vdimW1 vdimW2 vdimW3=> mxf mxg.
by rewrite !rank_elim_seqmxE !rank_elimP.
Qed.

(*
Local Open Scope ring_scope.


Fixpoint exnullmx (m n:nat) : Matrix K :=
  if m is (S p) then
   let zp := (zeroes K p) in
     if n is (S q) then
       let zq := (zeroes K q) in
         cM 0 zq zp (exnullmx p q)
       else (eM K) (* m is 0 *)
  else (eM K).     (* n is 0 *)

Lemma Matrix2mat_zero : forall m n, Matrix2mat m n (exnullmx m n) =
(0:'M[K]_(m,n)).
Proof.
elim => [| m hm]; elim => [ | n hi] //=.
rewrite seq2rV_zeroes seq2cV_zeroes hm //=.
have -> : 0%:M = 0 :> 'M[K]_1 by apply/matrixP => i j; rewrite !mxE GRing.mul0rn.
by rewrite /block_mx !row_mx0 col_mx0.
Qed.

Lemma exnullmx_ok : forall m n, Mok (exnullmx m n) m n.
Proof.
elim => [ | m hi] [ | n] //=. 
by rewrite !size_nseq (hi n) !eqxx. 
Qed.

*)

(*
Lemma ex_dimHomologyE: (mult ex_mxf ex_mxg) = (exnullmx (vdim W1) (vdim W3)) -> 
   ex_dimHomology = \dim (Homology (LinearApp (Matrix2mat (vdim W1) (vdim W2) ex_mxf))
                                   (LinearApp (Matrix2mat (vdim W2) (vdim W3) ex_mxg))).
Proof. 
move => nilpotency_mx; rewrite ex_dimHomology_dimHomologyrank dimHomologyrankE //=.
move : nilpotency_mx.
rewrite -{1}(m2m_ok2 Mokf) -{1}(m2m_ok2 Mokg) !vdimW1 !vdimW2 !vdimW3.
move : Sv1 => [x1 H1]; move : Sv2 => [x2 H2]; move : Sv3 => [x3 H3].
rewrite -H1 -H2 -H3 -multE.
move => H.
have H4 : (Matrix2mat (1 + x1) (1 + x2) ex_mxf *m Matrix2mat (1 + x2) (1 + x3) ex_mxg) =
          (Matrix2mat (1 + x1) (1 + x3) (exnullmx (1 + x1) (1 + x3))).
  by rewrite -H m2m_ok.
by rewrite H4 Matrix2mat_zero.
Qed.
*)

End Ex_Homology_Def.


(* Some examples *)



(* The first example is the torus *)
(*
Require Import boolF2.

Definition F2 := [cringType bool of bool].

Definition d0_ex1 := [::] : seqmatrix F2.
Definition d1_ex1 := [::[::false;false;false]] : seqmatrix F2.
Definition d2_ex1 := [::[::true;true];
                        [::true;true];
                        [::true;true]] : seqmatrix F2.
Definition d3_ex1 := [::] : seqmatrix F2.

Definition H0_ex1 := (ex_Betti_rank 0 1 3 d0_ex1 d1_ex1).
Definition H1_ex1 := (ex_Betti_rank 1 3 2 d1_ex1 d2_ex1).
Definition H2_ex1 := (ex_Betti_rank 3 2 0 d2_ex1 d3_ex1).

Eval vm_compute in H0_ex1. (* 1 connected component *)
Eval vm_compute in H1_ex1. (* 2 holes *)
Eval vm_compute in H2_ex1. (* 1 cavity *)

(* The second example comes from a triangle attached to an empty triangle *)

Definition d0_ex2 := [::] : seqmatrix F2.
Definition d1_ex2 := [::[::1;1;0;0;0];
                        [::1;0;1;1;0];
                        [::0;1;1;0;1];
                        [::0;0;0;1;1]] : seqmatrix F2.
Definition d2_ex2 := [::[::1;1;0;0]] : seqmatrix F2.

Definition H0_ex2 := (ex_Betti_rank 0 4 5 d0_ex2 d1_ex2). 
Definition H1_ex2 := (ex_Betti_rank 4 5 1 d1_ex2 d2_ex2).

Eval vm_compute in H0_ex2. (* 1 connected component *)
Eval vm_compute in H1_ex2. (* 1 hole *)

(* This example comes from an empty triangle *)

Definition d0_ex3 := [::] : seqmatrix F2.
Definition d1_ex3 := [::[::1;1;0];
                        [::1;0;1];
                        [::0;1;1]] : seqmatrix F2.
Definition d2_ex3 := [::] : seqmatrix F2.

Definition H0_ex3 := (ex_Betti_rank 0 3 3 d0_ex3 d1_ex3).
Definition H1_ex3 := (ex_Betti_rank 3 3 0 d1_ex3 d2_ex3).

Eval vm_compute in H0_ex3. (* 1 connected component *)
Eval vm_compute in H1_ex3. (* 1 hole *)



*)