(* We want to prove that the product of two consecutive incidence matrices *)
(* of a simplicial complex is equal to the null matrix.                    *)
Require Import ssreflect ssrbool ssrfun eqtype ssrnat seq div choice fintype.
Require Import finfun bigop finset ssralg perm zmodp matrix fingroup.
Require Import path vector seqmatrix boolF2 homology.

Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.

(* Compatibility layer for ssreflect 1.3 *)

Section compatibility.

Variable T : finType.
Implicit Types (a x : T) (A B C D : {set T}) (pA pB : pred T).

Lemma cardsD A B : #|A :\: B| = (#|A| - #|A :&: B|)%N.
Proof. by rewrite -(cardsID B A) addKn. Qed.

Lemma cardsDS A B : B \subset A -> #|A :\: B| = (#|A| - #|B|)%N.
Proof. by rewrite cardsD => /setIidPr->. Qed.

Definition powerset D := [set A : {set T} | A \subset D].

Lemma powersetE A B : (A \in powerset B) = (A \subset B).
Proof. by rewrite inE. Qed.

Lemma powerset0 : powerset set0 = [set set0] :> {set {set T}}.
Proof. by apply/setP=> A; rewrite !inE subset0. Qed.

Variable x : T.

Fixpoint rem s := if s is y :: t then (if y == x then t else y :: rem t) else s.

Lemma rem_filter s : uniq s -> rem s = filter (predC1 x) s.
Proof.
elim: s => //=. move => y s IHs /andP[not_s_y /IHs->].
by case: eqP => //= <-; apply/esym/all_filterP; rewrite all_predC has_pred1.
Qed.

Lemma rem_subseq s : subseq (rem s) s.
Proof.
elim: s => //= y s IHs; rewrite eq_sym.
by case: ifP => _; [exact: subseq_cons | rewrite eqxx].
Qed.

Lemma mem_rem_uniq s : uniq s -> rem s =i [predD1 s & x].
Proof. by move/rem_filter=> -> y; rewrite mem_filter. Qed.

End compatibility.

Section simplicial_complexes.

(* V is the vertex set *)
Variable V : finType.

(* A simplex over V is a finite subset of V *)
Definition simplex := {set V}.

Definition face_op (S : simplex) (x : V) := S :\ x.

Definition boundary (S : simplex) := (face_op S) @: S.

Lemma face_op_inj: forall (S : simplex),
  {in S, injective (face_op S)}.
Proof.
move=> S x Hx y HB.
have Hbx: #|S :\: (face_op S x)| == 1%N.
  rewrite cardsDS ; last exact: subD1set.
  by rewrite (cardsD1 x) -(add0n #|_ :\ _|) Hx add0n -addnBA // subnn.
case/cards1P:Hbx => t Ht.
have Hx2: x \in (S :\: (face_op S x)).
  by rewrite /face_op !inE eqxx Hx.
case Hy: (y \in S) ; last first.
  rewrite /face_op in HB.
  move:(ltnSn #|S :\ x|) ; move:(cardsD1 y S).
  rewrite (cardsD1 x) HB Hx Hy add1n add0n => Habs.
  by rewrite -{1}Habs ltnn.
have Hy2: y \in (S :\: (face_op S y)).
  by rewrite /face_op !inE eqxx Hy.
by rewrite -HB Ht !in_set1 in Hx2 Hy2 ; rewrite (eqP Hx2) (eqP Hy2).
Qed.

Lemma face_op_inj2: forall (S : simplex),
  {in S &, injective (face_op S)}.
Proof.
by move=> S x y Hx _ ; exact: (face_op_inj Hx).
Qed.

Lemma boundaryP: forall (S : simplex) (B : simplex),
  reflect (B \subset S /\ #|S| = #|B|.+1) (B \in boundary S).
Proof.
move=> S B ; apply:introP => H.
  case/imsetP: H => x Hx HB.
  split ; first by rewrite HB ; exact:subD1set.
  by rewrite (cardsD1 x) Hx add1n HB.
case=> H_sub H_card.
have: B \in boundary S ; last by rewrite (negbTE H).
apply/imsetP ; case:(pickP [pred x in (S :\: B)]) => [x Hx | H0].
  rewrite !inE in Hx ; case/andP:Hx => Hx1 Hx2.
  exists x => // ; apply/setP ; apply:subset_cardP.
    by apply/eqP ; rewrite -eqSS -H_card (cardsD1 x) Hx2 add1n.
  apply/subsetP => y Hy ; rewrite !inE (subsetP H_sub _ Hy) andbT.
  case Heq: (y == x) => // ; by rewrite -(eqP Heq) Hy in Hx1.
by move:(cardsDS H_sub) ; rewrite H_card subSnn (eq_card H0) card0.
Qed.

(* Simplicial complex definifition *)
Definition simplicial_complex (c : {set simplex}) :=
    forall x, x \in c -> forall y : simplex, y \subset x ->  y \in c.

(* utiliser bigcup_sup *)

Definition create_sc (s : seq simplex) : {set simplex} :=
  \bigcup_(sp <- s) powerset sp.

Lemma create_sc_correct : forall s, simplicial_complex (create_sc s).
move=> s x; rewrite /create_sc bigcup_seq.
case/bigcupP=> si Hsis Hxsi y Hy; apply/bigcupP; exists si=> //.
by rewrite powersetE; apply:(subset_trans Hy); rewrite -powersetE.
Qed.

Definition facet (c : seq simplex) (x : simplex) :=
   [set y : simplex | (x \subset y) && (y \in c)] == [set x].

Definition facets (c : seq simplex) := filter (facet c) c.

(* Definition of the incidence matrix of two enumerated finite sets of *)
(* simplices *)

Section incidence_matrix.

Open Local Scope ring_scope.

Variables m n : nat.

(* Top and Left are two enumerated finite sets of simplices *)
Variables Left Top : seq simplex.

(* x0 is the trivial simplex *)
Definition x0 := set0 : simplex.

(* We are going to work with coefficients in the ring Z2 *)
(* From the incidence function we can define the incidence matrix *)
Definition incidenceMatrix :=
  \matrix_(i < m, j < n)
    if nth set0 Left i \in boundary (nth set0 Top j) then 1 else 0:'F_2.

End incidence_matrix.

Section nth_incidence_mx.

(* We are going to define the incidence matrix of the set of simplices c *)
(* in dimension n using the above definition *)

Variable c: {set simplex}.
Variable n:nat.

(* The Top set is the set of n-simplices of c and the Left set is the  *) 
(* set of (n-1)-simplices of c *)

Definition n_1_simplices := if n==0 then set0 else [set x in c | #|x|==n] : {set simplex}.
Definition n_simplices := [set x in c | #|x|==n.+1] : {set simplex}.


Definition V1 := matrix_vectType [ringType of 'F_2] 1%N #|n_1_simplices|.
Definition V2 := matrix_vectType [ringType of 'F_2] 1%N #|n_simplices|.

(* Using the incidenceMatrix function we define the incidence matrix of *)
(* the set of simplices c in dimension n *)
Definition incidence_mx_n :=
  incidenceMatrix (Vector.dim V1) (Vector.dim V2) (enum n_1_simplices) (enum n_simplices).

End nth_incidence_mx.

End simplicial_complexes.

(* Statement of the theorem *)

Section incidence_matrices_sc_product.

Open Local Scope ring_scope.

Import GRing.Theory.

Lemma incidence_mulmx_seq (V:finType) (n:nat)
  (sc: {set (simplex V)}) s1 s2 s3 :
    uniq s1 -> uniq s2 -> uniq s3 ->
    s1 =i n_1_simplices sc n ->
    s2 =i n_simplices sc n ->
    s3 =i n_simplices sc n.+1 ->
    simplicial_complex sc ->
    (incidenceMatrix (Vector.dim (V1 sc n)) (Vector.dim (V2 sc n)) s1 s2) *m
    (incidenceMatrix (Vector.dim (V1 sc n.+1)) (Vector.dim (V2 sc n.+1)) s2 s3) = 0.
Proof.
move=> Hs1' Hs2' Hs3' Hs1 Hs2 Hs3 Hsc; apply/matrixP; rewrite /Vector.dim /=.
rewrite !mul1n -(eq_card Hs1) -(eq_card Hs2) -(eq_card Hs3) (card_uniqP Hs1').
rewrite (card_uniqP Hs2') (card_uniqP Hs3'); move=> i j; rewrite !mxE.
pose F x := if (nth set0 s1 i \in boundary x) &&
  (x \in boundary (nth set0 s3 j)) then 1 else 0:'F_2.
move:(mem_nth set0 (ltn_ord i)) (mem_nth set0 (ltn_ord j)).
rewrite Hs1 Hs3 !inE (eq_bigr (fun k => F (nth set0 s2 (val k)))); last first.
  by move=> k _; rewrite !mxE /F; case:ifP; case:ifP; rewrite ?mul0r ?mulr0.
case H0 : (n==0%N); first by rewrite /n_1_simplices H0 inE.
rewrite /n_1_simplices H0 inE.
case/andP => _ ; move/eqP=> Hi ; case/andP => Hj' ; move/eqP=> Hj.
rewrite -(big_mkord xpredT (fun k => F (nth set0 s2 k))).
rewrite -(big_nth set0 xpredT) /F big_uniq //=.
rewrite (bigID (mem (boundary (nth set0 s3 j)))) Monoid.mulmC /= big1; last first.
  by move=> x; case/andP=> [_ Hx]; rewrite (negbTE Hx) andbF.
have HB: {subset boundary (nth set0 s3 j) <= (n_simplices sc n)}.
  move=> x Hx; rewrite inE; case/boundaryP: Hx => Hx_sub Hx_card.
  by rewrite (Hsc _ Hj' _ Hx_sub) -eqSS -Hx_card Hj eqxx.
rewrite GRing.add0r (eq_bigl (mem (boundary (nth set0 s3 j)))) /=; last first.
  by move=> x; rewrite /= Hs2 -in_setI; move:x; apply/setP/setIidPr/subsetP.
pose G x := if nth set0 s1 i \in boundary x then 1 else 0:'F_2.
rewrite (eq_bigr G); last by move=> x ->; rewrite andbT.
rewrite big_imset; last exact:face_op_inj2.
rewrite /G (bigID (mem (nth set0 s1 i))) big1 /= ; last first.
  move=> x ; case/andP=> _ H; case:ifPn=> //=.
  by case/imsetP=> y _ Heq; rewrite Heq !inE eqxx andbF in H.
case Hij: (nth set0 s1 i \subset nth set0 s3 j) ; last first.
  rewrite big1 => [ |x]; first by rewrite add0r.
  case/andP => _ _ ; case:ifPn => //= ; case/boundaryP => H _.
  by rewrite (subset_trans H (subD1set _ _)) in Hij.
rewrite (eq_bigr (fun _ => 1)) ; last first.
  move=> x ; case/andP => H1 H2.
  suff: (nth set0 s1 i \in boundary (face_op (nth set0 s3 j) x)) => [->| ] //.
  apply/boundaryP ; split ; [apply/subsetP=> y Hy; rewrite !inE|apply/eqP].
    by rewrite ((subsetP Hij) y Hy); move:Hy H2; case:(y==x)/eqP => //= -> ->.
  by move:(cardsD1 x (nth set0 s3 j)); rewrite H1 Hi Hj addnC addn1 -eqSS => <-.
rewrite big_const (@eq_card _ _ [pred x in (nth set0 s3 j :\: nth set0 s1 i)]).
  rewrite (cardsDS Hij) Hi Hj -(addn1 n) -!addnS -{2}(addn0 n) subnDl //=
          addr0 add0r. 
  by apply:val_inj.
by move=> x; rewrite !inE andbC.
Qed.

Lemma incidence_matrices_sc_product (V:finType) (n:nat) (sc: {set (simplex V)}) :
  simplicial_complex sc ->
    (incidence_mx_n sc n) *m (incidence_mx_n sc (n.+1)) = 0.
Proof. by apply:incidence_mulmx_seq; rewrite ?enum_uniq //; exact:mem_enum. Qed.

Lemma incidence_matrices_sc_product_facets (V:finType) (n:nat)
  (ss: seq (simplex V)) : (n >= 1)%N ->
      (incidence_mx_n  (create_sc (facets ss)) n) *m
      (incidence_mx_n  (create_sc (facets ss)) (n.+1)) = 0.
Proof.
by rewrite incidence_matrices_sc_product //; apply: create_sc_correct.
Qed.

End incidence_matrices_sc_product.

Section executable_incidence_mx.

Local Open Scope ring_scope.

Variable V : finType.
Variable leT : rel V.

Hypothesis tr_leT : transitive leT.
Hypothesis irr_leT : irreflexive leT.

Variable faces : seq (seq V).
Variable n : nat.

Hypothesis Hfaces : all (sorted leT) faces.
Hypothesis Hfaces_uniq : all uniq faces.

Lemma setD1I (a:V) (A : {set V}) : a \notin A -> A :\ a = A.
Proof.
move=> H; apply/setP=> x; rewrite !inE; move:H; case:(x==a) / eqP=> // ->.
by case:(a \in A).
Qed.

Lemma setU1I (a:V) (A : {set V}) : a \in A -> a |: A = A.
Proof. by rewrite -sub1set=> H; rewrite (setUidPr H). Qed.

Lemma powerset_set1U (a:V) A :
  powerset (a |: A) = powerset A :|: [set a |: X | X in (powerset A)].
Proof.
apply/setP=> X; rewrite !inE; case HX: (X \subset A).
  by rewrite (subset_trans HX) // subsetU1.
rewrite -subDset; case HXa: (X :\ a \subset A).
  case Ha: (a \in X); last by rewrite (setD1I (negbT Ha)) HX in HXa.
  symmetry; apply/imsetP; exists (X :\ a); rewrite ?inE ?setD1K //.
case:imsetP=> //; case=> Y; rewrite inE=> HY HXY; case HaY: (a \in Y).
  by rewrite setU1I // in HXY; rewrite HXY HY in HX.
by rewrite HXY (setU1K (negbT HaY)) HY in HXa.
Qed.

Fixpoint powerset_seq (x : seq V) :=
  if x is h::t then let X := powerset_seq t in foldl (fun xs x => (h::x)::xs) X X
  else [::[::]].

(* For ssreflect 1.3 *)(*
Lemma imset_nil (T:finType) (f:T->T) : [set f x | x <- [::]] = set0.
Proof.
by apply/setP=> x; rewrite !inE Imset.imsetE !inE enum0.
Qed.*)

(* For ssreflect >= 1.4 *)

Lemma imset_nil (T:finType) (f:T->T) : [set f x | x in [::]] = set0.
Proof.
by apply/eqP; rewrite Imset.imsetE; unlock image_mem; rewrite enum0 /= eqxx.
Qed.


Lemma powerset_subseq (x : seq V) :
  all (subseq^~ x) (powerset_seq x).
Proof.
elim:x=> // a l IHl.
have H: forall acc, all (subseq^~ (a::l)) acc -> all (subseq^~ (a::l))
     (foldl (fun (xs : seq (seq V)) (x : seq V) => (a :: x) :: xs)
        acc (powerset_seq l)).
  move:IHl; elim: (powerset_seq l)=> // h s IHs IHl acc Hacc; apply:IHs=> //.
    by case/andP: IHl.
  by rewrite /= eqxx Hacc andbT; case/andP: IHl.
apply:H; apply/allP=> x Hx; apply:(subseq_trans _ (subseq_cons l a)).
exact:(allP IHl _ Hx).
Qed.

Definition bigcup_seq s (f : seq V -> seq (seq V)) :=
  foldl (fun xs x => (f x) ++ xs) [::] s.

Definition sc := undup (bigcup_seq faces powerset_seq).

Definition ex_face_op (S : seq V) x := rem x S.

Definition ex_boundary (S : seq V) :=
  foldl (fun xs x => ex_face_op S x :: xs) [::] S.

Definition ex_n_1_simplices := if n==0%N then nil else filter (fun s => size s == n) sc.
Definition ex_n_simplices := filter (fun s => size s == n.+1) sc.

Definition ex_incidence_mx  :=
  let ex_boundaries := map ex_boundary ex_n_simplices in
  map (fun x => map (fun y => if x \in y then true else false) ex_boundaries) ex_n_1_simplices.

Definition n_1_simplices_seq := [seq [set t in x] | x <- ex_n_1_simplices].
Definition n_simplices_seq := [seq [set t in x] | x <- ex_n_simplices].

Definition faces_set := [seq [set t in x] | x <- faces].
Definition sc_set := create_sc faces_set.

Lemma bigcup_seq_subseq s (f:seq V -> seq (seq V)) x :
  x \in bigcup_seq s f -> exists y, y \in s /\ x \in f y.
Proof.
have H: forall s acc, x \in
   foldl (fun (xs : seq (seq V)) (x0 : seq V) => f x0 ++ xs) acc s ->
   x \in acc \/ exists y : seq V, y \in s /\ x \in f y.
  elim=> [acc H|a l IHl acc H]; first by left.
  case:(IHl _ H).
    rewrite mem_cat; case/orP.
      by move=> Hx; right; exists a; split; rewrite ?inE ?eqxx.
    by move=> Hx; left.
  by case=> y [Hy Hxy]; right; exists y; split; rewrite ?inE //; apply/orP; right.
by move=> Hx; case:(H _ _ Hx).
Qed.

Lemma sc_subseq x : x \in sc -> exists y, y \in faces /\ subseq x y.
Proof.
rewrite mem_undup=> H; case (bigcup_seq_subseq H)=> y [Hy Hxy].
by exists y; split=> //; exact:(allP (powerset_subseq y)).
Qed.

Lemma sc_sorted : all (sorted leT) sc.
apply/allP=> x Hx; case:(sc_subseq Hx)=>y [Hy Hxy].
by apply:(subseq_sorted _ Hxy)=> //; exact:(allP Hfaces).
Qed.


Lemma sc_uniq : all uniq sc.
apply/allP=> x Hx; apply:(sorted_uniq tr_leT)=> //; exact:(allP sc_sorted).
Qed.

Lemma ex_n_1_simplices_sorted : all (sorted leT) ex_n_1_simplices.
Proof.
case H : (n==0%N); last by rewrite /ex_n_1_simplices H; apply/allP=> x; rewrite mem_filter; case/andP=> _; exact:(allP sc_sorted).
by rewrite /ex_n_1_simplices H.
Qed.

Lemma ex_n_simplices_sorted : all (sorted leT) ex_n_simplices.
Proof.
by apply/allP=> x; rewrite mem_filter; case/andP=> _; exact:(allP sc_sorted).
Qed.


Lemma eq_mem_map (T U : eqType) (f:T->U) (s s' :seq T) : s =i s' ->
  [seq f x | x <- s] =i [seq f x | x <- s'].
Proof.
move=> H x; apply/mapP; case:ifP.
  by case/mapP=> x' Hx' Hx; exists x'=> //; rewrite H.
move=> Habs [x' Hx' Hx]; move:Habs; rewrite Hx.
by have->//:(f x' \in [seq f x0 |  x0 <- s']); apply:map_f; rewrite -H.
Qed.

Lemma eq_mem_imset (T U : finType) (f:T->U) (s s' : pred T) :
  s =i s' -> [set f x | x in s] =i [set f x | x in s'].
Proof.
move=>H x; rewrite Imset.imsetE !inE; apply:eq_mem_map=> t.
by rewrite !mem_enum.
Qed.

Lemma imset_cons (T U : finType) (f:T->U) x s :
  [set f t | t in x :: s] = f x |: [set f t | t in s].
Proof.
apply/setP=> t.
have H: x :: s =i x |: [set u in s] by move=> u; rewrite !inE.
rewrite (eq_mem_imset f H) imsetU1 !inE; congr orb; apply: eq_mem_imset=> u.
by rewrite inE.
Qed.

Lemma powerset_seqP s :
  powerset [set t in s] =i [seq [set t in x] | x <- powerset_seq s].
Proof.
elim:s=> [x | h s IHs x]; first by rewrite powerset0 !inE.
have <-: forall s acc,
  (x \in [set h |: X | X in [seq [set t in x] | x <- s]])
    || (x \in [seq [set t in x] | x <- acc]) =
  (x \in [seq [set t in x] |
   x <- foldl (fun (xs : seq (seq V)) (x1 : seq V) => (h :: x1) :: xs) acc s]).
  elim=> [acc| a l IHl acc]; first by rewrite imset_nil inE.
  by rewrite -IHl /= !inE set_cons imset_cons !inE orbA [x in _ = x || _]orbC.
rewrite set_cons powerset_set1U in_setU IHs orbC.
by congr orb; exact:eq_mem_imset.
Qed.

Lemma bigcup_seqP : \bigcup_(sp <- faces_set) powerset sp =i
  [seq [set t in x] |  x <- bigcup_seq faces powerset_seq].
Proof.
move=> x.
(* rewrite /faces_set.
rewrite /bigcup_seq.*)
have <-: forall acc,
  ((x \in \bigcup_(sp <- [seq [set t in x1] |  x1 <- faces]) powerset sp)
     || (x \in [seq [set t in x] | x <- acc])) =
   (x \in [seq [set t in x1]
       |  x1 <- foldl (fun (xs : seq (seq V)) (x1 : seq V) =>
            powerset_seq x1 ++ xs) acc faces]).
  elim:faces=> [acc| h s IHs acc]; first by rewrite big_nil !inE.
  rewrite /bigcup_seq /= -IHs big_cons in_setU powerset_seqP map_cat mem_cat.
  by rewrite orbA [x in _ = x || _]orbC.
by rewrite orbF.
Qed.

Lemma scP : sc_set =i [seq [set t in x] | x <- sc].
Proof.
by move=> x; rewrite (eq_mem_map _ (mem_undup _)); exact:bigcup_seqP.
Qed.

Lemma sc_strong_uniq : {in sc &, injective (fun x => [set t in x])}.
Proof.
move=> x y Hx Hy /= /setP Hxy; apply:(eq_sorted_irr tr_leT)=> //.
  + exact:(allP sc_sorted x Hx).
  + exact:(allP sc_sorted y Hy).
  by move=> t; move:(Hxy t); rewrite !inE.
Qed.

Lemma n_simplices_strong_uniq :
  {in ex_n_simplices &, injective (fun x => [set t in x])}.
Proof.
move=> x y; rewrite !mem_filter; case/andP=> _ Hx; case/andP=> _ Hy.
exact:sc_strong_uniq.
Qed.

Lemma n_1_simplices_strong_uniq :
  {in ex_n_1_simplices &, injective (fun x => [set t in x])}.
Proof.
case H : (n==0%N); rewrite /ex_n_1_simplices H; last first.
  move=> x y; rewrite !mem_filter; case/andP=> _ Hx; case/andP=> _ Hy.
  exact:sc_strong_uniq.
by done.
Qed.


Definition IM := incidenceMatrix (Vector.dim (V1 sc_set n)) (Vector.dim (V2 sc_set n)) n_1_simplices_seq n_simplices_seq.

Lemma uniq_ex_n_1_simplices : uniq ex_n_1_simplices.
Proof. 
by case H : (n==0%N); rewrite /ex_n_1_simplices H; last by apply:filter_uniq; exact:undup_uniq.
Qed.

Lemma uniq_ex_n_simplices : uniq ex_n_simplices.
Proof. apply:filter_uniq; exact:undup_uniq. Qed.

Lemma uniq_n_1_simplices_seq : uniq n_1_simplices_seq.
Proof.
by rewrite (map_inj_in_uniq n_1_simplices_strong_uniq) uniq_ex_n_1_simplices.
Qed.

Lemma uniq_n_simplices_seq : uniq n_simplices_seq.
Proof.
by rewrite (map_inj_in_uniq n_simplices_strong_uniq) uniq_ex_n_simplices.
Qed.

Lemma n_1_simplices_seqE :
  n_1_simplices_seq =i n_1_simplices sc_set n.
Proof.
case H0 : (n==0%N); rewrite /n_1_simplices H0.
  by rewrite /n_1_simplices_seq /ex_n_1_simplices H0 /=; move => x;
  rewrite in_set0 in_nil.
move=> x; rewrite /n_1_simplices_seq /ex_n_1_simplices H0 inE scP; apply/mapP; case:ifP.
  case/andP; case/mapP=> x' Hx' Hxx' Hx; exists x'=> //.
  rewrite mem_filter Hx'; rewrite Hxx' cardsE in Hx.
  by rewrite -(card_uniqP ((allP sc_uniq) x' Hx')) Hx.
move=> H; case=> x'; rewrite mem_filter; case/andP=> Hsx' Hx' Hxx'.
move:Hsx' H.
rewrite -(card_uniqP ((allP sc_uniq) x' Hx')) -[#|x'|]cardsE -Hxx' => ->.
rewrite Hxx'.
have->//:[set t in x'] \in [seq [set t in x0] |  x0 <- sc].
exact:map_f.
Qed.

Lemma n_simplices_seqE :
  n_simplices_seq =i n_simplices sc_set n.
Proof.
move=> x; rewrite inE scP; apply/mapP; case:ifP.
  case/andP; case/mapP=> x' Hx' Hxx' Hx; exists x'=> //.
  rewrite mem_filter Hx'; rewrite Hxx' cardsE in Hx.
  by rewrite -(card_uniqP ((allP sc_uniq) x' Hx')) Hx.
move=> H; case=> x'; rewrite mem_filter; case/andP=> Hsx' Hx' Hxx'.
move:Hsx' H.
rewrite -(card_uniqP ((allP sc_uniq) x' Hx')) -[#|x'|]cardsE -Hxx' => ->.
rewrite Hxx'.
have->//:[set t in x'] \in [seq [set t in x0] |  x0 <- sc].
exact:map_f.
Qed.

Lemma n_1_simplicesE :
   size ex_n_1_simplices = #|n_1_simplices sc_set n|.
Proof.
transitivity (size n_1_simplices_seq); first by rewrite size_map.
move/card_uniqP:uniq_n_1_simplices_seq => <-; apply:eq_card.
exact:n_1_simplices_seqE.
Qed.

Lemma n_simplicesE:
  size ex_n_simplices = #|n_simplices sc_set n|.
Proof.
transitivity (size n_simplices_seq); first by rewrite size_map.
move/card_uniqP:uniq_n_simplices_seq => <-; apply:eq_card.
exact:n_simplices_seqE.
Qed.

Lemma face_opP s x : uniq s ->
  face_op [set t in s] x =i [set t in ex_face_op s x].
Proof. by move=> Hs t; rewrite !inE mem_rem_uniq // inE. Qed.

Lemma ex_boundaryP s :
  uniq s ->
  boundary [set t in s] =i [seq [set t in x] | x <- ex_boundary s].
Proof.
have H: forall acc s' x, uniq s'->
  (x \in [set face_op [set t in s'] x | x in [set t in s]])
    || (x \in [seq [set t in x1] |  x1 <- acc]) =
   (x \in [seq [set t in x1]
      |  x1 <- foldl (fun (xs : seq (seq V)) (x1 : V) => 
         ex_face_op s' x1 :: xs) acc s]).
  elim:s;first by move=> acc s' x' Hs'; rewrite /boundary imset0 !inE.
  move=> h s IHs acc s' x' Hs'.
  rewrite set_cons imsetU !inE -IHs //= inE [x in _ = x]orbA.
  rewrite [x in _ = x || _]orbC !imset_set1 inE.
  by move/setP:(face_opP h Hs')=> ->.
by move=> Hs x; rewrite -H // orbF.
Qed.

Lemma ex_boundary_sorted s :
  sorted leT s ->
  all (sorted leT) (ex_boundary s).
Proof.
move=> Hs.
have->//:forall s s' acc, sorted leT s' -> all (sorted leT) acc ->
    all (sorted leT) (foldl (fun xs x => ex_face_op s' x :: xs) acc s).
elim=> // a l IHl acc s' Hal Hacc; apply:IHl=> //.
rewrite /= Hacc andbT /ex_face_op; apply:(subseq_sorted tr_leT _ Hal).
exact:rem_subseq.
Qed.

(*
Lemma IME : ex_incidence_mx = seqmx_of_mx IM.
Proof.
rewrite seqmx_of_funE /vdim /= !mul1n -n_1_simplicesE -n_simplicesE.
apply:eq_mkseqmx_ord=> i j; rewrite !(nth_map [::]) // ex_boundaryP.
  case:ifP=> H; first by rewrite (map_f (fun (u:seq V) => [set t \in u]) H).
  case:ifP=> //; case/mapP=> x Hx Hxi.
  have: x = nth [::] ex_n_1_simplices i.
    apply:(eq_sorted_irr tr_leT irr_leT).
      + apply:(allP (ex_boundary_sorted _) _ Hx).
        by apply:(allP (ex_n_simplices_sorted)); exact:mem_nth.
      + by apply:(allP ex_n_1_simplices_sorted); exact:mem_nth.
      by move=> t; move/setP:Hxi=> Hxi; move:(Hxi t); rewrite !inE.
  by move=> H'; rewrite H' H in Hx.
apply:(allP sc_uniq); move:(mem_nth [::] (ltn_ord j)); rewrite mem_filter.
by case/andP.
Qed.
*)



Lemma IME : ex_incidence_mx = 
 (@seqmx_of_mx [ringType of 'F_2] bool_cringType (Vector.dim (V1 sc_set n))
             (Vector.dim (V2 sc_set n)) IM).
Proof.
apply/seqmxP; split=> [|i|i j].
    by rewrite size_map /Vector.dim /= mul1n -n_1_simplicesE.
  rewrite /Vector.dim /= mul1n -n_1_simplicesE=> Hi.
  by rewrite /rowseqmx (nth_map [::]) // !size_map mul1n -n_simplicesE.
rewrite mxE; move: i j; rewrite /fun_of_seqmx /rowseqmx.
rewrite /Vector.dim /= !mul1n -n_1_simplicesE -n_simplicesE=> i j.
rewrite !(nth_map [::]) //; last by rewrite size_map.
rewrite ex_boundaryP.
case:ifP=> H; first by rewrite (map_f (fun (u:seq V) => [set t in u]) H).
  case:ifP=> //; case/mapP=> x Hx Hxi.
  have: x = nth [::] ex_n_1_simplices i.
    apply:(eq_sorted_irr tr_leT irr_leT).
      + apply:(allP (ex_boundary_sorted _) _ Hx).
        by apply:(allP (ex_n_simplices_sorted)); exact:mem_nth.
      + by apply:(allP ex_n_1_simplices_sorted); exact:mem_nth.
      by move=> t; move/setP:Hxi=> Hxi; move:(Hxi t); rewrite !inE.
  by move=> H'; rewrite H' H in Hx.
apply:(allP sc_uniq); move:(mem_nth [::] (ltn_ord j)); rewrite mem_filter.
by case/andP.
Qed.

End executable_incidence_mx.


Section homology_sc.

Variable (V: finType).


Definition Homology_sc_n (sc : {set simplex V}) (n : nat) :=
  Homology (Vector.Hom (incidence_mx_n sc n.+1)^T) 
           (Vector.Hom (incidence_mx_n sc n)^T).

Definition Betti_sc_n (sc : {set simplex V}) (n : nat) :=
  Betti (Vector.Hom (incidence_mx_n sc n.+1)^T) 
           (Vector.Hom (incidence_mx_n sc n)^T).

Definition ex_Betti_sc (sc : seq (seq V)) n :=
  ex_Betti_rank (size (ex_n_1_simplices sc n))
                (size (ex_n_simplices sc n))
                (size (ex_n_simplices sc n.+1))
                (ex_incidence_mx sc n)
                (ex_incidence_mx sc n.+1).

End homology_sc.

