(*****************************************************************************)
(*                        Persitent Homology Groups                          *)
(*****************************************************************************)
Require Import ssreflect ssrfun ssrbool eqtype ssrnat seq choice fintype bigop.
Require Import finfun tuple ssralg matrix mxalgebra zmodp seqmatrix vector.
Require Import cssralg rank.

Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.


(*****************************************************************************

Definition of homology groups:

Let V1, V2 and V3 be vector spaces over a field K and f: V1 -> V2, g: V2 -> V3
be linear maps such that gf = 0 then the homology group of (f,g), denoted by
H_(f,g) is the quotient ker(g)/im(f).

*****************************************************************************)

Section PHomologyDefinition.

Variable (F : fieldType).
Variables (V1 V2 V3 V4 : vectType F)
          (f : 'Hom(V1,V2)) (g : 'Hom(V3,V4))
          (i : 'Hom(V1 ,V4)).

Hypothesis (i_inj : injective i).

Definition PHomology := ((i @: (lker f)) :\: ((limg g) :&: (i @: (lker f))))%VS.

Section PHomologyDimension.

Definition PBetti := \dim PHomology.

Lemma addn_equal_subn (m n p : nat) : (m + n = p -> m = p - n)%N.
Proof. by move=> H; rewrite -H addnK. Qed.

Lemma dim_ker : \dim (lker f) =  Vector.dim V1 - \dim (limg f).
Proof.
have : \dim ((@fullv _ V1) :&: lker f) + \dim (f @: (@fullv _ V1)) = \dim (@fullv _ V1).
  by rewrite limg_ker_dim . 
have H1 : \dim ((@fullv _ V1) :&: lker f) = \dim (lker f).
  by move : (subvf (lker f)); move/capv_idPl; rewrite capvC => ->.  
have H2 : \dim (f @: (@fullv _ V1)) = \dim (limg f) by done.
rewrite H1 dimvf => H.
exact: addn_equal_subn.
Qed.

Lemma PBettiE : PBetti = Vector.dim V1 - \dim (limg f) - \dim ((limg g) :&: (i @: (lker f))).
Proof.
rewrite /PBetti /PHomology.
have -> :  \dim (i @: lker f :\: limg g :&: i @: lker f) =
           \dim (i @: lker f) -  \dim (i @: lker f :&: (limg g :&: i @: lker f)).
  apply: addn_equal_subn.
  by rewrite addnC dimv_cap_compl.
rewrite limg_dim_eq; last first.
  have -> : lker i = (0%VS) by apply/eqP/lker0P.
  by rewrite capv0.
rewrite dim_ker.
suff : (limg g :&: i @: lker f <= i @: lker f)%VS.
  by move/capv_idPl; rewrite capvC => ->.
by rewrite capvSr.
Qed.

Lemma PBettiE_2 :
  PBetti = Vector.dim V1 - \dim (limg f) - (\dim (limg g) +
              (Vector.dim V1 - \dim (limg f)) -  \dim ((limg g) + (i @: (lker f)))).
Proof.
rewrite PBettiE.
suff -> : \dim (limg g :&: i @: lker f) = (\dim (limg g) +
          (Vector.dim V1 - \dim (limg f)) - \dim (limg g + i @: lker f))=> //.
have -> : \dim (limg g :&: i @: lker f) = \dim (limg g) + \dim (i@: lker f) - \dim (limg g + i @: lker f).
  by rewrite -dimv_sum_cap -[X in _ - X]addn0 subnDl subn0.
rewrite -dim_ker (@limg_dim_eq _ _ _ i (lker f)) //. 
have -> : lker i = (0%VS)%R by apply/eqP/lker0P.
by rewrite capv0.
Qed.

End PHomologyDimension.

End PHomologyDefinition.

Section kernel.

Local Open Scope ring_scope.

Variable (K : fieldType).
Variable (m n l : nat) (M : 'M[K]_(m,n)) (L : 'M[K]_(l,m)).

Hypothesis sub_kermx2P : forall p (X : 'M[K]_(p,m)), reflect (X *m M = 0) (X <= L)%MS.

Lemma eqmx_kermx : (L :=: kermx M)%MS.
Proof.
apply/eqmxP/andP.
split; first by apply/sub_kermxP/sub_kermx2P.
by apply/sub_kermx2P; rewrite mulmx_ker.
Qed.

End kernel.

Section kermx2.

Variable K : fieldType.

Local Open Scope ring_scope.

Require Import kermx.

Lemma eqmx_ker m n (M : 'M[K]_(m,n)) : (ker M :=: kermx M)%MS.
Proof. by apply/eqmx_kermx => p X; apply/sub_kerP. Qed.

End kermx2.


Section PHomologyRank.

Variable K : fieldType.
Variable (v1 v2 v3 v4 : nat).

Definition V1 := (matrix_vectType K 1 v1).
Definition V2 := (matrix_vectType K 1 v2).
Definition V3 := (matrix_vectType K 1 v3).
Definition V4 := (matrix_vectType K 1 v4).

Variable (mxf: 'M[K]_(Vector.dim V1,Vector.dim V2)) (mxg : 'M[K]_(Vector.dim V3,Vector.dim V4))
         (mxi : 'M[K]_(Vector.dim V1,Vector.dim V4)).

Local Open Scope ring_scope.

Lemma vdimV1: Vector.dim V1 = v1.
Proof. by rewrite /Vector.dim /= mul1n. Qed.

Definition PBetti_rank :=
   (v1 - \rank mxf - (\rank mxg + (v1 - \rank mxf) -
   \rank (col_mx mxg ((ker mxf) *m mxi))))%N.

Lemma PBetti_rankE : injective (Vector.Hom mxi) ->
  PBetti_rank = PBetti (Vector.Hom mxf) (Vector.Hom mxg) (Vector.Hom mxi).
Proof.
move=> hinj.
rewrite PBettiE_2 /PBetti_rank /ker // -addsmxE.
rewrite {10 11}vdimV1. 
have -> : \dim (limg (Vector.Hom mxf)) = \rank mxf.
  by rewrite /dimv /lfun_img unlock //= genmx1 mul1mx mxrank_gen.
have -> : \dim (limg (Vector.Hom mxg)) = \rank mxg.
  by rewrite /dimv /lfun_img unlock //= genmx1 mul1mx mxrank_gen.
have H : (kermx mxf :=: <<kermx mxf>>)%MS
  by apply/eqmx_sym/genmxE.
move : (eqmx_trans (eqmx_ker mxf) H) => H1.
have H2 : (mxg + ker mxf *m mxi :=: mxg + <<kermx mxf>> *m mxi)%MS.
 apply/adds_eqmx; first by apply:eqmx_refl.
 by apply : eqmxMr.
suff -> : \rank (mxg + projT2 (ker_dep mxf) *m mxi) =
     \dim (limg (Vector.Hom mxg) + Vector.Hom mxi @: lker (Vector.Hom mxf)).
  by done.
rewrite H2.
rewrite /dimv /lfun_img. 
have -> : (Vector.InternalTheory.vs2mx
            (locked_with lfun_img_key (lfun_img_def (K:=K)) V3 V4
               (Vector.Hom mxg) fullv +
             locked_with lfun_img_key (lfun_img_def (K:=K)) V1 V4
               (Vector.Hom mxi) (lker (Vector.Hom mxf)))) = 
          <<(<<mxg>> + <<<<kermx mxf>> *m mxi>>)%MS>>%MS.
 by rewrite unlock //= genmx1 mul1mx.
by rewrite -mxrank_gen genmx_adds mxrank_gen.
Qed.

End PHomologyRank.

Section Ex_PHomology.

Variable (K : fieldType).
Variable (CK : cunitRingType K).
Variable (v1 v2 v3 v4 : nat).

Hypothesis v1_0 : v1 > 0.
Hypothesis v2_0 : v2 > 0.
Hypothesis v3_0 : v3 > 0.
Hypothesis v4_0 : v4 > 0.

Definition W1 := (matrix_vectType K 1 v1).
Definition W2 := (matrix_vectType K 1 v2).
Definition W3 := (matrix_vectType K 1 v3).
Definition W4 := (matrix_vectType K 1 v4).

Lemma vdimW1: Vector.dim W1 = v1.
Proof. by rewrite /W1 /Vector.dim /= mul1n. Qed.

Lemma vdimW2: Vector.dim W2 = v2.
Proof. by rewrite /W2 /Vector.dim /= mul1n. Qed.

Lemma vdimW3: Vector.dim W3 = v3.
Proof. by rewrite /W3 /Vector.dim /= mul1n. Qed.

Lemma vdimW4: Vector.dim W4 = v4.
Proof. by rewrite /W4 /Vector.dim /= mul1n. Qed.

Let rank := rank_elim_seqmx.

(*
(* TODO: Correct definition! *)
(* The size is projT1 (ker M) but this is not nice as we have to make
   an executable function computing this... *)
Definition ker_seqmx m n (M : seqmatrix CK) : seqmatrix CK :=
  seqmx0 CK (m - rank m n M) m.

Lemma ker_dep_seqmxE : forall m n (M : 'M_(m, n)),
  ker_seqmx m n (seqmx_of_mx CK M) = seqmx_of_mx CK (ker M).
Proof.
move=> m n M.
by admit.
Qed.
*)

Definition ex_PBetti_rank (mxf mxg mxi : seqmatrix CK) :=
  let rf := rank v1 v2 mxf in
  let rg := rank v3 v4 mxg in
  v1 - rf - (rg + (v1 - rf) -
  rank (v3 + size (ker_seqmx v1 v2 mxf)) v4
       (col_seqmx mxg (mulseqmx v1 v4 (ker_seqmx v1 v2 mxf) mxi))).


Lemma ex_PBetti_rank_PBetti_rank_E :
  forall (mxf: 'M[K]_(Vector.dim W1, Vector.dim W2))
         (mxg : 'M[K]_(Vector.dim W3, Vector.dim W4))
         (mxi : 'M[K]_(Vector.dim W1, Vector.dim W4)),
   ex_PBetti_rank (seqmx_of_mx CK mxf) (seqmx_of_mx CK mxg) (seqmx_of_mx CK mxi) =
   PBetti_rank mxf mxg mxi.
Proof.
rewrite /ex_PBetti_rank /PBetti_rank vdimW1 vdimW2 vdimW3 vdimW4 /rank=> mxf mxg mxi.
rewrite !rank_elim_seqmxE !rank_elimP.
rewrite -ker_seqmxE.
case: v1 v1_0 mxf mxi => // n _ mxf mxi.
by rewrite mulseqmxE col_seqmxE size_seqmx rank_elim_seqmxE rank_elimP.
Qed.



End Ex_PHomology.


